"""
Models and Enums for the database.
"""

from enum import Enum, auto

from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

def get_or_create(session, model, **kwargs):
    """
    Returns an instance of a database model, creates one and returns that if not found.
    """

    instance = session.query(model).filter_by(**kwargs).first()

    if not instance:
        instance = model(**kwargs)
        session.add(instance)
        session.commit()

    return instance


class ServerChannels(Base):
    """
    Database table for server channels.
    These are discord server channels to be included when sharing news from certain categories.
    """

    __tablename__ = 'server_channels'

    # Columns
    id = Column(Integer, primary_key=True, autoincrement=True)
    server_id = Column(Integer, nullable=False)
    channel_id = Column(Integer, nullable=False)
    news_category_id = Column(Integer, nullable=False)
    active = Column(Integer, nullable=False, default=True)
    news_articles = relationship('NewsArticles', cascade='all, delete')


class NewsArticles(Base):
    """
    Database table for news articles.
    This is not for information about news articles, but rather for keeping track of where
    certain articles have been sent to, so that they arent shared twice in the same place.
    """

    __tablename__ = 'news_articles'

    # Columns
    id = Column(Integer, primary_key=True, autoincrement=True)
    url = Column(String, nullable=False)
    server_channel_id = Column(Integer, ForeignKey('server_channels.id'), nullable=False)


class NewsCategories(Base):
    """
    Database table for news categories.
    Used to categorise news articles to RSS feeds. Contains default entries that should not be
    modified or added to.
    """

    __tablename__ = 'news_categories'

    # Columns
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, unique=True, nullable=False)


class DefaultNewsCategories(Enum):
    """
    Enum for news categories. Used to create default data in the database table of the same name.
    """

    ALL = auto()
    WORLD = auto()
    UK = auto()
    NORTH_AMERICA = auto()
    ENTERTAINMENT = auto()
    BUSINESS = auto()
    TECH = auto()
    SCIENCE = auto()
    TOP_STORIES = auto()
