"""
Extension for the `ErrorCog` cog.
Loading this file via `commands.Bot.load_extension` will add the `ErrorCog` cog to the bot.
"""

import logging

from discord import app_commands, Interaction
from discord.ext import commands
from discord.errors import InteractionResponded

log = logging.getLogger(__name__)


class ErrorCog(commands.Cog):
    """
    Error handling cog.
    Discordpy has problems with error handling, this cog corrects
    """

    default_err_msg = "I'm sorry, but I've encountered an " \
                      "error while processing your command."

    def __init__(self, bot):
        super().__init__()
        self.bot = bot

        # Register the error handler
        bot.tree.error(coro=self._dispatch_to_app_command_handler)

    def trace_error(self, error: Exception):
        log.error(f"{type(error).__name__} {error}")
        raise error

    async def _dispatch_to_app_command_handler(
        self,
        inter: Interaction,
        error: app_commands.AppCommandError
    ):
        """
        Dispatches the error to the app command handler.
        """

        self.bot.dispatch("app_command_error", inter, error)

    async def _respond_to_interaction(self, inter: Interaction) -> bool:
        """
        Respond to an interaction with an error message.
        """

        try:
            await inter.response.send_message(
                self.default_err_msg,
                ephemeral=True
            )
        except InteractionResponded:
            log.debug("Interaction already responded to.")
            return

    @commands.Cog.listener("on_app_command_error")
    async def get_app_command_error(
        self,
        inter: Interaction,
        error: app_commands.AppCommandError
    ):
        """
        Handles the application command error and responds with the appropriate error message.
        """

        try:
            # Send the default error message and create an edit
            # shorthand to add more details to the message once
            # we've figured out what the error is.
            log.error(error.with_traceback(None))
            await self._respond_to_interaction(inter)
            edit = lambda x: inter.edit_original_response(content=x)

            raise error

        except app_commands.CommandInvokeError as _err:

            # The interaction has already been responded to.
            if isinstance(_err.original, InteractionResponded):
                await edit(_err.original)
                return

            # Some other error occurred while invoking the command.
            await edit(f"`{type(_err.original).__name__}` : {_err.original}")

        except app_commands.CheckFailure as _err:

            # The command is still on cooldown.
            if isinstance(_err, app_commands.CommandOnCooldown):
                await edit(
                    f"Woah, slow down! This command is on cooldown, " \
                    f"wait `{str(_err).split(' ')[7]}` !"
                )
                return

            if isinstance(_err, app_commands.MissingPermissions):
                await edit(
                    "You don't have the required permissions to " \
                    "run this command!"
                )
                return

            if isinstance(_err, app_commands.BotMissingPermissions):
                await edit(
                    "I don't have the required permissions to " \
                    "run this command! Please ask an admin to " \
                    "grant me the required permissions."
                )
                return

            # A different check has failed.
            await edit(f"`{type(_err).__name__}` : {_err}")
        
        except app_commands.CommandNotFound:

            # The command could not be found.
            await edit(
                f"I couldn't find the command you were looking for... "
                "\nThis is probably a discord bug related to " \
                "desynchronization between my commands and discord's " \
                "servers. Please try again later."
            )

        except Exception as _err: 
            # Caught here:
                # app_commands.TransformerError
                # app_commands.CommandLimitReached
                # app_commands.CommandAlreadyRegistered
                # app_commands.CommandSignatureMismatch

            self.trace_error(_err)


async def setup(bot):
    """
    Setup function for this extension.
    Adds the `ErrorCog` cog to the bot.
    """

    cog = ErrorCog(bot)
    await bot.add_cog(cog)
    log.info(f"Added {cog.__class__.__name__} cog")
