"""
Extension for the `NewsCog` cog.
Loading this file via `commands.Bot.load_extension` will add the `NewsCog` cog to the bot.
"""

import asyncio
import logging
from datetime import datetime

import aiohttp
import discord
from discord import app_commands, Interaction
from discord.ext import commands, tasks
from sqlalchemy import select, insert, delete, and_
from bbc_feeds import news as news_api, feedparser
from bs4 import BeautifulSoup as bs4

from db.db import DatabaseManager
from db.models import ServerChannels, DefaultNewsCategories, NewsArticles

log = logging.getLogger(__name__)

category_choices = [
    app_commands.Choice(name=category.name.replace("_", " "), value=category.value)
    for category in DefaultNewsCategories
]


class NewsStoryType:
    """
    Type hinting class for news stories.
    """

    id: str
    title: str
    title_detail: str
    summary: str
    summary_detail: str
    link: str
    links: list[str]
    guidislink: str
    published: str
    published_parsed: datetime


class NewsCog(commands.Cog):
    """
    News cog.
    Delivers embeds of news articles to discord channels.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot
        self.news_task.start()

    def fetch_articles(self, category_id: int) -> list[NewsStoryType]:
        """
        Fetch the latest news articles from the category matching the given `category_id`.

        Arguments
        ---------
        category_id : int
            The ID of the category to fetch news articles from.

        Raises
        ------
        ValueError
            The given `category_id` doesn't match any known categories.
        """

        n = news_api()

        # IMPORTANT:
        # The order of items in this list matters,
        # Items follow the order of items in the Enum: DefaultNewsCategories
        category_methods = [
            n.all,
            n.world,
            n.uk,
            n.north_america,
            n.entertainment,
            n.business,
            n.tech,
            n.science,
            n.top_stories
        ]

        try:
            return category_methods[category_id - 1](limit=1)
        except IndexError as err:
            raise ValueError(f"Invalid category_id: {category_id}") from err

    async def story_to_embed(self, story: NewsStoryType, category_name: str) -> discord.Embed:
        """
        Returns a discord.Embed object representing the given story.

        Parameters
        ----------
        story : NewsStoryType
            Parsed details on the news story.

        Returns
        -------
        discord.Embed
            A `discord.Embed` object updated with information on the news story.
        """

        # Fetch web data for the thumbnail
        async with aiohttp.ClientSession() as session:
            async with session.get(story.link) as response:
                html = await response.text()

        # Parse the thumbnail for the news story
        soup = bs4(html, "html.parser")
        image_src = soup.select_one("meta[property='og:image']").get("content")

        category_name = category_name.replace("_", " ").title()

        embed = discord.Embed(
            colour=discord.Colour.from_str("#FFFFFF"),
            title=story.title,
            description=story.summary,
            url=story.link,
        )
        embed.set_image(url=image_src)
        embed.set_author(name=f"BBC News  •  {category_name}")

        return embed

    async def followup_with_articles(self, inter: Interaction, category: app_commands.Choice[int]):
        """
        Collects articles and follows up an interaction with embeds for these articles.

        Arguments
        ---------
        inter : discord.Interaction
            The interaction between the bot and the user.

        category : app_commands.Choice[int]
            The category chosen by the user, represented as a `Choice` object.
        """

        stories = self.fetch_articles(category.value)

        if not stories:
            await inter.followup.send("No articles found")
            return

        for story in stories:
            embed = await self.story_to_embed(story, category.name)
            await inter.followup.send(embed=embed)

    async def _get_or_fetch_channel(self, channel_id: int) -> discord.TextChannel:
        """
        Returns a `discord.TextChannel` object based on the given `channel_id`.
        Will try to return from cache if present, otherwise will make an API call for the channel.
        """

        channel = self.bot.get_channel(channel_id)
        return channel or await self.bot.fetch_channel(channel_id)

    @tasks.loop(minutes=5)
    async def news_task(self):
        """
        Task that checks for the latest news and shares it to discord.
        """

        log.info("Doing news task.")

        async with DatabaseManager() as database:
            for category in DefaultNewsCategories:

                # Find channels that accept this category of news
                whereclause = and_(ServerChannels.news_category_id == category.value)
                query = select(ServerChannels).where(whereclause)
                result = await database.session.execute(query)
                server_channels = result.scalars().all()

                if not server_channels:
                    continue

                stories = self.fetch_articles(category.value)
                if not stories:
                    continue

                story = stories[0]
                embed = None

                for item in server_channels:

                    channel_id = item.channel_id

                    # Check the article hasn't already been shared with this channel.
                    query = select(NewsArticles).where(and_(NewsArticles.server_channel_id == channel_id, NewsArticles.url == story.link))
                    result = await database.session.execute(query)
                    existing_items = result.scalars().all()

                    if existing_items:
                        log_string = (
                            "Existing items not sent"
                            f"\nchannel id: {channel_id}"
                            f"\narticle url: {story.link}"
                            f"\ncategory: {category.name}\n"
                        )
                        log.info(log_string)
                        continue

                    log_string = (
                            "Items sent"
                            f"\nchannel id: {channel_id}"
                            f"\narticle url: {story.link}"
                            f"\ncategory: {category.name}\n"
                        )
                    log.info(log_string)

                    # Add this article as shown to this channel, prevents it from being shown here again.
                    query = insert(NewsArticles).values(server_channel_id=channel_id, url=story.link)
                    await database.session.execute(query)

                    embed = embed or await self.story_to_embed(story, category.name)

                    channel = await self._get_or_fetch_channel(channel_id)
                    await channel.send(embed=embed)

            await database.session.commit()  # commit will terminate the connection for some reason.

    base_group = app_commands.Group(name="bbc", description="BBC News related commands")
    channels_group = app_commands.Group(parent=base_group, name="channels", description="Channel commands")

    @base_group.command(name="news")
    @app_commands.choices(category=category_choices)
    async def get_news(self, inter: Interaction, category: app_commands.Choice[int]):
        """
        Get the latest article from BBC news.
        """

        await inter.response.defer()
        await self.followup_with_articles(inter, category)

    # @base_group.command(name="ping-me")
    # async def ping_me(self, inter: Interaction):
    #     """
    #     Ping you when a new news article has been posted. Use again to remove ping.
    #     """

    #     await inter.response.send_message("response OK")

    @channels_group.command(name="add-category")
    @app_commands.choices(category=category_choices)
    async def add_category(self, inter: Interaction, channel: discord.TextChannel, category: app_commands.Choice[int]):
        """
        Assign a category to this channel. Articles under this category will be sent here.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = insert(ServerChannels).values(server_id=inter.guild_id, channel_id=channel.id, news_category_id=category.value)
            await database.session.execute(query)
            await database.session.commit()

        category_name = category.name.replace("_", " ").title()
        embed = discord.Embed(
            title=f"Category Added  •  {category_name}",
            description=f"{channel.mention} will now receive news from **{category_name}**",
            colour=discord.Colour.from_str("#FFFFFF")
        )

        await inter.followup.send(embed=embed)

    @channels_group.command(name="del-category")
    @app_commands.choices(category=category_choices)
    async def delete_category(self, inter: Interaction, channel: discord.TextChannel, category: app_commands.Choice[int]):
        """
        Remove a category from this channel.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            whereclause = and_(
                ServerChannels.channel_id == channel.id,
                ServerChannels.news_category_id == category.value
            )
            query = delete(ServerChannels).where(whereclause)
            await database.session.execute(query)
            await database.session.commit()

        category_name = category.name.replace("_", " ").title()
        embed = discord.Embed(
            title=f"Category Removed  •  {category_name}",
            description=f"{channel.mention} will no longer receive news from **{category_name}**",
            colour=discord.Colour.from_str("#FFFFFF")
        )

        await inter.followup.send(embed=embed)

    @channels_group.command(name="lst-category")
    async def list_category(self, inter: Interaction):
        """
        List the categories assigned to this channel.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = select(ServerChannels)
            result = await database.session.execute(query)
            server_channels = result.scalars().all()

            output = ""

            for item in server_channels:
                if item.server_id != inter.guild_id:
                    continue

                channel = inter.guild.get_channel(item.channel_id)
                category_name = DefaultNewsCategories(item.news_category_id).name.replace("_", " ")

                output += f"{channel.mention} - {category_name}\n"

        output = output or "No categories set"
        await inter.followup.send(output)


async def setup(bot):
    """
    Setup function for this extension.
    Adds the `ErrorCog` cog to the bot.
    """

    cog = NewsCog(bot)
    await bot.add_cog(cog)
    log.info(f"Added {cog.__class__.__name__} cog")
